/*-------------- Telecommunications & Signal Processing Lab ---------------
                             McGill University

Routine:
  long int CPcopyChan (AFILE *AFp[], const long int StartF[], int Nifiles,
                       long int Nframe, long int MaxNframe, AFILE *AFpO)

Purpose:
  Copy audio file samples

Description:
  This routine copies input audio file(s) to an output audio file.  The input
  channels are spread out over one or more files.  All input channels appear
  in the output file.  The number of output sample frames can be specified in a
  number of ways.  Let the number of frames as determined by the file limits
  be Nframe and the maximum number of frames be MaxNframe.  Designate the
  number of frames to the end-of-file be EoF.  The combinations are as follows.
                                      MaxNframe
                             specified        unspecified

    Nframe   specified    min(Nframe, MaxNframe)  Nframe
           unspecified    min(MaxNframe, EoF)     EoF

Parameters:
  <-  long int CPcopyChan
      Number of sample frames written
   -> AFILE *AFp[]
      Array of Nifiles input audio file pointers
   -> const long int StartF[]
      Start frames for the input files
   -> int Nifiles
      Number of input files
   -> long int Nframe
      Number of output frames
   -> long int MaxNframe
      Maximum number of frames to be written to the output file
   -> AFILE *AFpO
      Output audio file pointer
 
Author / revision:
  P. Kabal  Copyright (C) 1999
  $Revision: 1.19 $  $Date: 1999/06/05 00:58:31 $

-------------------------------------------------------------------------*/

static char rcsid[] = "$Id: CPcopyChan.c 1.19 1999/06/05 AFsp-v4r3 $";

#include <assert.h>
#include <limits.h>

#include <libtsp.h>
#include "CopyAudio.h"

#define MINV(a, b)	(((a) < (b)) ? (a) : (b))
#define MAXV(a, b)	(((a) > (b)) ? (a) : (b))
#define ICEILV(n, m)	(((n) + ((m) - 1)) / (m))	/* int n,m >= 0 */

#define BFSIZE	5120

static long int
CP_copy1 (AFILE *AFpI, long int StartF, long int Nframe, long int MaxNframe,
	  AFILE *AFpO);
static long int
CP_copyNa (AFILE *AFp[], const long int StartF[], int Nifiles, long int Nframe,
	   AFILE *AFpO);
static long int
CP_copyNb (AFILE *AFp[], const long int StartF[], int Nifiles, long int Nframe,
	   AFILE *AFpO);


long int
CPcopyChan (AFILE *AFp[], const long int StartF[], int Nifiles,
	    long int Nframe, long int MaxNframe, AFILE *AFpO)

{
  long int NchanI, NchanMax, Nfw;
  int j, Ns;

/*
   Cases:
   (1) single input file - the copying is sample by sample without regard to
       channels
   (N) multiple input files - the input channels have to be interleaved in the
       output.  The number of samples must be known.
       (a) The number of channels is such that one or more record from the
           file with the largest number of channels can fit into the buffer
	   along with the output record.
       (b) The number of channels is large.  The data from a single input file
           may have to be read in chunks.  Each chunk is immediately sent to
           the output.
*/

/* Determine the total number of input channels and the maximum number of
   input channels to be read from any file */
  NchanMax = 0;
  NchanI = 0;
  for (j = 0; j < Nifiles; ++j) {
    NchanMax = MAXV (NchanMax, AFp[j]->Nchan);
    NchanI += AFp[j]->Nchan;
  }
  assert (NchanI == AFpO->Nchan);

/* Single input file */
  if (Nifiles == 1)

    Nfw = CP_copy1 (AFp[0], StartF[0], Nframe, MaxNframe, AFpO);

/* Multiple input files */
  else {
    Ns = (int) (BFSIZE / (NchanMax + AFpO->Nchan));
    if (Ns >= 1)
      Nfw = CP_copyNa (AFp, StartF, Nifiles, Nframe, AFpO);
    else
      Nfw = CP_copyNb (AFp, StartF, Nifiles, Nframe, AFpO);
  }

  return Nfw;
}

/* Copy one file */


static long int
CP_copy1 (AFILE *AFpI, long int StartF, long int Nframe, long int MaxNframe,
	  AFILE *AFpO)

{
  float Fbuf[BFSIZE];
  long int offr, offs, Nrem, Nj;
  int eof, Nr, Nv;

/* Copy sample by sample from the input file to the output file */
/* *** offsets and counts are sample offsets, not frame values *** */
  Nj = AFpI->Nchan;

  eof = (Nframe == AF_NFRAME_UNDEF);
  if (MaxNframe == AF_NFRAME_UNDEF) {
    if (eof)
      Nrem = LONG_MAX;
    else
      Nrem = Nj * Nframe;
  }
  else {
    if (eof)
      Nrem = Nj * MaxNframe;
    else
      Nrem = Nj * MINV (Nframe, MaxNframe);
  }

  offr = 0L;
  while (Nrem > 0L) {
    Nv = (int) MINV (Nrem, BFSIZE);
    offs = offr + Nj * StartF;
    Nr = AFreadData (AFpI, offs, Fbuf, Nv);

    if (Nr < Nv && eof) {
      offs = offr + Nr;
      Nrem = Nj * ICEILV (offs, Nj) - offr;
      Nv = MINV (Nrem, Nv);
    }
    Nrem -= Nv;
    offr += Nv;

    AFwriteData (AFpO, Fbuf, Nv);
  }
  assert (Nj * (offr/Nj) == offr);

  return offr/Nj;
}

/* Copy multiple files, number of channels small */


static long int
CP_copyNa (AFILE *AFp[], const long int StartF[], int Nifiles, long int Nframe,
	   AFILE *AFpO)

{
  int i, j, k, n, Ns, Nj, NO, Nfr;
  long int offr, offs, Nrem, NchanMax;
  float Fbuf[BFSIZE];
  float *Fbufi, *Fbufo;

  assert (Nframe != AF_NFRAME_UNDEF);

/* Determine the maximum number of input channels to be read from any file */
  NchanMax = 0L;
  for (j = 0; j < Nifiles; ++j)
    NchanMax = MAXV (NchanMax, AFp[j]->Nchan);

/* Split the buffer space up for the input and output buffers */
  NO = (int) AFpO->Nchan;
  Ns = (int) (BFSIZE / (NchanMax + NO));
  assert (Ns > 0);

  Fbufi = Fbuf;
  Fbufo = Fbuf + ((int) NchanMax) * Ns;

  offr = 0L;
  Nrem = Nframe;
  while (Nrem > 0) {

    Nfr = (int) MINV (Nrem, Ns);
    n = 0;
    for (j = 0; j < Nifiles; ++j) {
      Nj = (int) AFp[j]->Nchan;
      offs = Nj * (offr + StartF[j]);
      AFreadData (AFp[j], offs, Fbufi, Nfr * Nj);
      for (k = 0; k < Nj; ++k) {
	for (i = 0; i < Nfr; ++i)
	  Fbufo[i*NO+n+k] = Fbufi[i*Nj+k];
      }
      n += Nj;
    }

    /* Write the samples to the output file */
    AFwriteData (AFpO, Fbufo, Nfr * NO);
    offr += Nfr;
    Nrem -= Nfr;
  }

  return Nframe;
}

/* Copy multiple files, number of channels large */


static long int
CP_copyNb (AFILE *AFp[], const long int StartF[], int Nifiles, long int Nframe,
	   AFILE *AFpO)

{
  int j, Nc;
  long int offr, offs, l, Nj;
  float Fbuf[BFSIZE];

  assert (Nframe != AF_NFRAME_UNDEF);

/* Large number of channels, handle one input file at a time */
  for (offr = 0L; offr < Nframe; ++offr) {
    for (j = 0; j < Nifiles; ++j) {
      Nj = AFp[j]->Nchan;
      for (l = 0L; l < Nj; l += Nc) {
	Nc = (int) MINV (Nj - l, BFSIZE);
	offs = Nj * (offr + StartF[j]) + l;
	AFreadData (AFp[j], offs, Fbuf, Nc);
	AFwriteData (AFpO, Fbuf, Nc);
      }
    }
  }

  return Nframe;
}
