/*-------------- Telecommunications & Signal Processing Lab ---------------
                             McGill University

Routine:
  void ADsetOFormat (int audio_fd, double Sfreq, int Nchan)

Purpose:
  Set the output conversion parameters for the Sun audio device

Description:
  This routine sets the output conversion parameters for the Sun audio device.
  The sampling rate is set to the requested value, or one of a number of
  standard values if the standard rate within 5 % of the requested value.  The
  default output data type is integer if the audio device supports it.  The
  fallback output data type is mu-law.

Parameters:
   -> int audio_fd
      File descriptor for the audio device
   -> double Sfreq
      Sampling frequency
   -> int Nchan
      Number of channels

Author / revision:
  P. Kabal  Copyright (C) 1999
  $Revision: 1.21 $  $Date: 1999/06/14 16:19:59 $

-------------------------------------------------------------------------*/

static char rcsid[] = "$Id: ADsetOFormat.c 1.21 1999/06/14 AFsp-v4r3 $";

#include <assert.h>
#include <errno.h>
#include <stropts.h>		/* stream ioctl commands */
#include <unistd.h>		/* ioctl */

#include <libtsp.h>
#include <libtsp/AFpar.h>
#include <AD.h>

#define ABSV(x)		(((x) < 0) ? -(x) : (x))
#define NELEM(array)	((sizeof array) / (sizeof array[0]))

#define NFreq		(NELEM (StdFreq))
#define PGM		(UTgetProg ())

static double StdFreq[] = {
  8000.,
  9600.,
  11025.,
  16000.,
  18900.,
  22050.,
  32000.,
  37800.,
  44100.,
  48000.
};

static int
AD_setFormat (int audio_fd, unsigned int Encoding, int Nchan);
static int
AD_setFreq (int audio_fd, double Sfreq);


void
ADsetOFormat (int audio_fd, double Sfreq, int Nchan)

{
  int i;
  int iost;
  unsigned int Encoding;

  assert (Sfreq > 0.0);

/* ==================== Sampling frequency */
  iost = AD_setFreq (audio_fd, Sfreq);
  if (iost == -1) {
    /* Incompatible sampling frequency */
    /*
      See if the desired sampling frequency is close to one of a number of
      standard frequencies (those for the dbri and audiocs interfaces)
    */
    for (i = 0; i < NFreq; ++i) {
      if (ABSV (Sfreq - StdFreq[i]) <= SFREQ_TOL * Sfreq)
	break;
    }

    /* Try setting the sampling frequency again */
    if (i < NFreq)
      iost = AD_setFreq (audio_fd, StdFreq[i]);

    if (iost == -1)
      UThalt ("%s: %s: \"%g\"", PGM, ADM_UnsSFreq, Sfreq);

    UTwarn (ADMF_SFreqChange, Sfreq, StdFreq[i]);
  }

/* ==================== Data type */
  if (Nchan > 2)
    UThalt ("%s: %s: \"%ld\"", PGM, ADM_XNChan, Nchan);

  iost = AD_setFormat (audio_fd, AUDIO_ENCODING_LINEAR, Nchan);
  if (iost == -1) {

    /* Fallback data type */
    Encoding = AUDIO_ENCODING_ULAW;

    iost = AD_setFormat (audio_fd, Encoding, Nchan);
    if (iost == -1) {

      /* Fallback to single channel */
      iost = AD_setFormat (audio_fd, Encoding, 1);
      if (iost == -1)
	UTerror ("%s: %s", PGM, ADM_DataNChan);
    }
  }

  return;
}

/*
  Set the sampling rate.
  iost:  zero,     no error
         non-zero, invalid value
*/


static int
AD_setFreq (int audio_fd, double Sfreq)

{
   audio_info_t aif;
   int iost;

   AUDIO_INITINFO (&aif);
   aif.play.sample_rate = (int) (Sfreq + 0.5);
   iost = ioctl (audio_fd, AUDIO_SETINFO, &aif);

   if (iost == -1 && errno != EINVAL)
     UTerror ("%s: %s", PGM, ADM_SetErr);	/* Generic error */

   return iost;
}

/*
  Set the data type
  iost:  zero,      no error
         non-zero, invalid value
*/


static int
AD_setFormat (int audio_fd, unsigned int Encoding, int Nchan)

{
   audio_info_t aif;
   int iost;

   AUDIO_INITINFO (&aif);
   if (Encoding == AUDIO_ENCODING_LINEAR)
     aif.play.precision = 16;
   else if (Encoding == AUDIO_ENCODING_ULAW)
     aif.play.precision = 8;
   else
     assert (0);	/* Unsupported encoding */

   aif.play.encoding = Encoding; 
   aif.play.channels = Nchan;
   iost = ioctl (audio_fd, AUDIO_SETINFO, &aif);

   if (iost == -1 && errno != EINVAL)
     UTerror ("%s: %s", PGM, ADM_SetErr);	/* Generic error */

   return iost;
}
