/*-------------- Telecommunications & Signal Processing Lab ---------------
                             McGill University

Routine:
  void AFprAFpar (AFILE *AFp, const char Fname[], FILE *fpinfo)

Purpose:
  Print information about an audio file

Description:
  This routine optionally prints header information for an audio file.

Parameters:
   -> AFILE *AFp
      Audio file pointer for the audio file.  If AFp is NULL, this routine
      is a no-op.
   -> const char Fname[]
      File name.  If the file is stdin or stdout, this string is not used.
   -> FILE *fpinfo
      File pointer for printing the audio file identification information.  If
      fpinfo is NULL, no information is printed.

Author / revision:
  P. Kabal  Copyright (C) 1999
  $Revision: 1.46 $  $Date: 1999/06/04 20:07:54 $

-------------------------------------------------------------------------*/

static char rcsid [] = "$Id: AFprAFpar.c 1.46 1999/06/04 AFsp-v4r3 $";

#include <assert.h>

#include <libtsp.h>
#include <libtsp/nucleus.h>
#include <libtsp/AFheader.h>
#include <libtsp/AFmsg.h>
#define AF_DATA_TYPE_NAMES
#define AF_FILE_TYPE_NAMES
#include <libtsp/AFpar.h>

static const char *
AF_findRec (const char *Id[], const struct AF_info *Hinfo);


void
AFprAFpar (AFILE *AFp, const char Fname[], FILE *fpinfo)

{
  const char *ft;
  const char *Datetime, *Title;
  char FullName[FILENAME_MAX];
  static const char *IDtitle[] = {"title:", "display_name:", "name:",
				  "user_comment:", NULL};
  static const char *IDdate[] = {"date:", "recording_date:", "creation_date:",
				 NULL};

  if (AFp == NULL)
    return;

  assert (AFp->Op == FO_RO || AFp->Op == FO_WO);
  assert (AFp->Format > 0 && AFp->Format < NFD);
  assert (AFp->Ftype > 0 && AFp->Ftype < NFT);

/* Print the header information */
  if (fpinfo != NULL) {

    /* Type of file */
    if (AFp->Ftype == FT_NH && AFopt_NHpar.Start == 0L)
      ft = AFM_Headerless;
    else
      ft = AF_FTN[AFp->Ftype];
    fprintf (fpinfo, AFMF_File, ft);

    /* File name */
    if (AFp->fp == stdin)
      fprintf (fpinfo, AFMF_Name, "<stdin>");
    else if (AFp->fp == stdout)
      fprintf (fpinfo, AFMF_Name, "<stdout>");
    else {
      FLfullName (Fname, FullName);
      fprintf (fpinfo, AFMF_Name, FullName);
    }

    /* Title */
    Title = AF_findRec (IDtitle, &AFp->Hinfo);
    if (Title != NULL)
      fprintf (fpinfo, AFMF_Desc, STstrDots (Title, 56));

    /* Number of samples or frames */
    if (AFp->Op == FO_RO) {
      if (AFp->Nsamp == AF_NSAMP_UNDEF)
	fprintf (fpinfo, AFMF_NumSampUnk);
      else {
	if (AFp->Nchan > 1)
	  fprintf (fpinfo, AFMF_NumFrame, AFp->Nsamp / AFp->Nchan);
	else
	  fprintf (fpinfo, AFMF_NumSamp, AFp->Nsamp);
      }

      /* Duration */
      if (AFp->Nsamp != AF_NSAMP_UNDEF && AFp->Sfreq > 0.0)
	fprintf (fpinfo, AFMF_Duration, (AFp->Nsamp / AFp->Nchan)/AFp->Sfreq);

      /* Date */
      Datetime = AF_findRec (IDdate, &AFp->Hinfo);
      if (Datetime == NULL) {
	Datetime = "";
	if (AFp->fp != stdin)
	  Datetime = FLfileDate (AFp->fp, 3);
      }
      fprintf (fpinfo, AFMF_Date, Datetime);
    }

    /* Sampling frequency */
    fprintf (fpinfo, AFMF_SFreq, AFp->Sfreq);

    /* Number of channels and data format */
    fprintf (fpinfo, AFMF_NumChan, AFp->Nchan);
    fprintf (fpinfo, AFMF_Format, AF_DTN[AFp->Format]);
  }

  return;
}

/* Search for named records in an AFsp information structure */


static const char *
AF_findRec (const char *Id[], const struct AF_info *Hinfo)

{
  int i;
  const char *p;

  p = NULL;
  for (i = 0; Id[i] != NULL; ++i) {
    p = AFgetHrec (Id[i], Hinfo);
    if (p != NULL)
      break;
  }

  if (p != NULL)
    p = STtrimIws (p);

  return p;
}
